==> mule-metadata-model-xml-1.1.7-sources.jar: org/mule/metadata/xml/api/XmlTypeLoader.java <==
/*
 * Copyright (c) MuleSoft, Inc.  All rights reserved.  http://www.mulesoft.com
 * The software in this package is published under the terms of the CPAL v1.0
 * license, a copy of which has been included with this distribution in the
 * LICENSE.txt file.
 */
package org.mule.metadata.xml.api;

import static java.util.stream.Collectors.joining;
import static org.mule.metadata.api.utils.MetadataTypeUtils.addTypeAlias;
import static org.mule.metadata.api.utils.MetadataTypeUtils.isNotNull;
import static org.mule.metadata.xml.internal.utils.SchemaHelper.getDefaultValue;

import org.mule.apache.xerces.impl.dv.xs.XSSimpleTypeDecl;
import org.mule.apache.xerces.impl.xs.XSComplexTypeDecl;
import org.mule.apache.xerces.impl.xs.XSElementDecl;
import org.mule.apache.xerces.xs.XSAnnotation;
import org.mule.apache.xerces.xs.XSAttributeDeclaration;
import org.mule.apache.xerces.xs.XSAttributeUse;
import org.mule.apache.xerces.xs.XSComplexTypeDefinition;
import org.mule.apache.xerces.xs.XSElementDeclaration;
import org.mule.apache.xerces.xs.XSModel;
import org.mule.apache.xerces.xs.XSModelGroup;
import org.mule.apache.xerces.xs.XSObjectList;
import org.mule.apache.xerces.xs.XSParticle;
import org.mule.apache.xerces.xs.XSTerm;
import org.mule.apache.xerces.xs.XSTypeDefinition;
import org.mule.apache.xerces.xs.XSWildcard;
import org.mule.metadata.api.TypeLoader;
import org.mule.metadata.api.builder.AttributeFieldTypeBuilder;
import org.mule.metadata.api.builder.BaseTypeBuilder;
import org.mule.metadata.api.builder.ObjectFieldTypeBuilder;
import org.mule.metadata.api.builder.ObjectKeyBuilder;
import org.mule.metadata.api.builder.ObjectTypeBuilder;
import org.mule.metadata.api.builder.TypeBuilder;
import org.mule.metadata.api.builder.UnionTypeBuilder;
import org.mule.metadata.api.model.MetadataFormat;
import org.mule.metadata.api.model.MetadataType;
import org.mule.metadata.xml.api.handler.HandlerManager;
import org.mule.metadata.xml.api.utils.SchemaHelper;

import javax.xml.namespace.QName;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Function;

import org.w3c.dom.TypeInfo;

public class XmlTypeLoader implements TypeLoader {

  public static final String TEXT_CONTENT_FIELD_NAME = "text()";
  public static MetadataFormat XML = MetadataFormat.XML;

  private HandlerManager handlerManager;
  private ModelFactory modelFactory;
  private Map<QName, TypeBuilder> types;

  /**
   * Creates a new instance based on a group of schemas.
   *
   * @param schemaCollector an {@link SchemaCollector} with the schemas.
   */
  public XmlTypeLoader(SchemaCollector schemaCollector) {
    this(ModelFactory.fromSchemas(schemaCollector.collect()));
  }

  /**
   * Creates a new instance based on an already constructed {@link ModelFactory}.
   *
   * @param modelFactory a {@link ModelFactory} instance.
   */
  public XmlTypeLoader(ModelFactory modelFactory) {
    this.modelFactory = modelFactory;
    this.types = new HashMap<>();
    this.handlerManager = new HandlerManager();
  }

  @Override
  public Optional<MetadataType> load(String typeIdentifier) {
    return load(typeIdentifier, null);
  }

  @Override
  public Optional<MetadataType> load(String typeIdentifier, String typeAlias) {
    return load(typeIdentifier != null ? QName.valueOf(typeIdentifier) : null, typeAlias);
  }

  public Optional<MetadataType> load(QName qName, String typeAlias) {
    try {
      final XSModel model = modelFactory.getModel();
      QName rootElement = qName == null ? modelFactory.getRootElementName().orElse(null) : qName;
      if (model != null) {

        final XSElementDeclaration elementDeclaration =
            model.getElementDeclaration(rootElement.getLocalPart(), rootElement.getNamespaceURI());

        if (elementDeclaration == null) {
          return Optional.empty();
        }
        final BaseTypeBuilder rootBuilder = BaseTypeBuilder.create(XML);
        // Create root object with the root name
        final ObjectTypeBuilder rootObject = rootBuilder.objectType();
        rootObject.id(rootElement.toString());
        modelFactory.getExample().ifPresent(rootObject::with);
        final ObjectFieldTypeBuilder field = rootObject.addField().required();
        final ObjectKeyBuilder key = field.key(rootElement);

        processAttributes(elementDeclaration, key);
        final BaseTypeBuilder valueBuilder = field.value();
        processValue(elementDeclaration, valueBuilder);
        addTypeAlias(rootObject, typeAlias);
        return Optional.ofNullable(rootBuilder.build());
      } else {
        return Optional.empty();
      }
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  private void processValue(XSElementDeclaration elementDeclaration, BaseTypeBuilder valueBuilder) {
    final XSTypeDefinition typeDefinition = getXsTypeDefinition(elementDeclaration);
    if (typeDefinition instanceof XSComplexTypeDecl) {
      final XSComplexTypeDecl complexType = (XSComplexTypeDecl) typeDefinition;
      final Optional<Short> groupType = getGroupType(complexType);
      if (groupType.isPresent() && groupType.get().equals(XSModelGroup.COMPOSITOR_CHOICE)) {
        processUnionType(valueBuilder, complexType);
      } else {
        if (XSComplexTypeDefinition.CONTENTTYPE_SIMPLE == complexType.getContentType()) {
          valueBuilder.stringType();
        } else {
          final ObjectTypeBuilder objectType = valueBuilder.objectType();
          if (XSComplexTypeDefinition.CONTENTTYPE_MIXED == complexType.getContentType()) {
            objectType.addField().key(TEXT_CONTENT_FIELD_NAME).value().stringType();
          }
          final Optional<QName> typeName = getTypeName(typeDefinition);
          if (typeName.isPresent()) {
            objectType.id(typeName.get().toString());
            types.put(typeName.get(), objectType);
          }
          processAnnotations(complexType, objectType);
          processElements(complexType, objectType);
        }
      }
    } else if (typeDefinition instanceof XSSimpleTypeDecl) {
      handlerManager.handle((XSSimpleTypeDecl) typeDefinition, getDefaultValue(elementDeclaration), valueBuilder);
    }
  }

  private XSTypeDefinition getXsTypeDefinition(XSElementDeclaration elementDeclaration) {
    if (elementDeclaration instanceof XSElementDecl) {
      org.mule.apache.xerces.xni.QName unresolved = ((XSElementDecl) elementDeclaration).fUnresolvedTypeName;
      if (unresolved != null) {
        XSTypeDefinition definition = modelFactory.getModel().getTypeDefinition(unresolved.localpart, unresolved.uri);
        if (definition != null) {
          return definition;
        }
      }
    }
    return elementDeclaration.getTypeDefinition();
  }

  private void processUnionType(BaseTypeBuilder typeBuilder, XSComplexTypeDecl complexType) {
    final UnionTypeBuilder unionType = typeBuilder.unionType();
    final Optional<QName> unionTypeName = getTypeName(complexType);
    unionTypeName.ifPresent(qName -> types.put(qName, unionType));
    final XSParticle particle = complexType.getParticle();
    if (particle != null) {
      final XSTerm term = particle.getTerm();
      if (term instanceof XSModelGroup) {
        final XSModelGroup xsModelGroup = (XSModelGroup) term;
        final XSObjectList particles = xsModelGroup.getParticles();
        for (Object xsObject : particles) {
          if (xsObject instanceof XSParticle) {
            final XSParticle part = (XSParticle) xsObject;
            final XSTerm element = part.getTerm();
            final ObjectTypeBuilder objectType = unionType.of().objectType();
            if (element instanceof XSElementDeclaration) {
              final XSElementDeclaration elementDeclaration = (XSElementDeclaration) element;
              final String name = element.getName();
              final String namespace = element.getNamespace();
              final ObjectFieldTypeBuilder field = objectType.addField();
              final ObjectKeyBuilder key = field.key(new QName(namespace, name));
              processAttributes(elementDeclaration, key);
              field.required(part.getMinOccurs() > 0);
              final XSTypeDefinition typeDefinition = ((XSElementDeclaration) element).getTypeDefinition();
              final Optional<QName> typeName = getTypeName(typeDefinition);
              if (typeName.isPresent() && types.containsKey(typeName.get())) {
                field.value(types.get(typeName.get()));
              } else {
                final BaseTypeBuilder fieldValue = field.value();
                processValue(elementDeclaration, fieldValue);
              }
            } else {
              processElements(part, objectType);
            }
          }
        }
      }
    }
  }

  private Optional<Short> getGroupType(XSComplexTypeDecl complexType) {
    final XSParticle particle = complexType.getParticle();
    if (particle != null) {
      final XSTerm term = particle.getTerm();
      if (term instanceof XSModelGroup) {
        return Optional.of(((XSModelGroup) term).getCompositor());
      }
    }
    return Optional.empty();
  }

  private void processAnnotations(XSComplexTypeDecl complexType, ObjectTypeBuilder objectType) {
    final XSObjectList annotations = complexType.getAnnotations();
    for (Object annotation : annotations) {
      final XSAnnotation xsAnnotation = (XSAnnotation) annotation;
      final Collection<SchemaHelper.XmlDoc> xmlDocs = SchemaHelper.getDocumentation(xsAnnotation.getAnnotationString());
      String content = xmlDocs.stream().map(doc -> doc.getContent()).collect(joining("\n "));
      Optional<String> language = xmlDocs.stream().map(doc -> doc.getLang().orElse(null)).filter(Objects::nonNull).findFirst();
      objectType.description(language.orElse(null), content);
    }
  }

  private void processAttributes(XSElementDeclaration elementDeclaration, ObjectKeyBuilder keyBuilder) {
    final XSTypeDefinition typeDefinition = elementDeclaration.getTypeDefinition();
    if (typeDefinition instanceof XSComplexTypeDecl) {
      final XSObjectList attributeUses = ((XSComplexTypeDefinition) typeDefinition).getAttributeUses();
      for (Object attributeUse : attributeUses) {
        if (attributeUse instanceof XSAttributeUse) {
          final XSAttributeUse xsAttributeUse = (XSAttributeUse) attributeUse;
          final XSAttributeDeclaration attrDeclaration = xsAttributeUse.getAttrDeclaration();
          final String localName = attrDeclaration.getName();
          final String namespace = attrDeclaration.getNamespace();
          final AttributeFieldTypeBuilder attributeField = keyBuilder.addAttribute();
          attributeField.name(new QName(namespace, localName));
          attributeField.required(xsAttributeUse.getRequired());
          handlerManager.handle((XSSimpleTypeDecl) attrDeclaration.getTypeDefinition(),
                                getDefaultValue(xsAttributeUse),
                                attributeField.value());
        }
      }
    }
  }

  private void processElements(XSComplexTypeDecl complexType, ObjectTypeBuilder objectType) {
    processElements(complexType.getParticle(), objectType);
  }

  private void processElements(XSParticle particle, ObjectTypeBuilder objectType) {
    processElements(particle, objectType, (part) -> part.getMinOccurs() > 0);
  }

  private void processElements(XSParticle particle, ObjectTypeBuilder objectType,
                               Function<XSParticle, Boolean> requiredSupplier) {
    if (particle == null) {
      return;
    }

    final XSTerm term = particle.getTerm();
    if (term instanceof XSModelGroup) {
      final XSModelGroup xsModelGroup = (XSModelGroup) term;
      objectType.ordered(xsModelGroup.getCompositor() == XSModelGroup.COMPOSITOR_SEQUENCE);
      final XSObjectList particles = xsModelGroup.getParticles();
      for (Object xsObject : particles) {
        if (xsObject instanceof XSParticle) {
          final XSParticle part = (XSParticle) xsObject;
          final XSTerm element = part.getTerm();
          if (element instanceof XSElementDeclaration) {
            final XSElementDeclaration xsElementDeclaration = (XSElementDeclaration) element;
            final String name = element.getName();
            final String namespace = element.getNamespace();
            final ObjectFieldTypeBuilder field = objectType.addField();
            processAttributes(xsElementDeclaration, field.key(new QName(namespace, name)));
            field.required(requiredSupplier.apply(part));
            field.repeated(isRepeated(part));

            Number min = null;
            Number max = null;

            if (part.getMinOccurs() > 1) {
              min = part.getMinOccurs();
            }

            if (part.getMaxOccurs() > 1) {
              max = part.getMaxOccurs();
            }

            if (isNotNull(min) || isNotNull(max)) {
              field.occurrence(min, max);
            }

            final XSTypeDefinition typeDefinition = ((XSElementDeclaration) element).getTypeDefinition();
            final Optional<QName> typeName = getTypeName(typeDefinition);
            if (typeName.isPresent() && types.containsKey(typeName.get())) {
              field.value(types.get(typeName.get()));
            } else {
              final BaseTypeBuilder fieldValue = field.value();
              processValue(xsElementDeclaration, fieldValue);
            }
          } else if (element instanceof XSWildcard) {
            if (((XSWildcard) element).getConstraintType() == XSWildcard.NSCONSTRAINT_ANY) {
              objectType.open();
            }
          } else if (element instanceof XSModelGroup) {
            processElements(part, objectType, (xsParticle) -> false);
          }
        }
      }
    }
  }

  private Optional<QName> getTypeName(XSTypeDefinition typeDefinition) {
    Optional<QName> typeQname = Optional.empty();
    if (typeDefinition instanceof TypeInfo) {
      final TypeInfo definition = (TypeInfo) typeDefinition;
      if (definition.getTypeName() != null) {
        typeQname = Optional.of(new QName(definition.getTypeNamespace(), definition.getTypeName()));
      }
    }
    return typeQname;
  }

  private boolean isRepeated(XSParticle part) {
    return part.getMaxOccursUnbounded() || part.getMaxOccurs() > 1 || part.getMinOccurs() > 1;
  }

}

==> mule-metadata-model-xml-1.2.4-sources.jar: org/mule/metadata/xml/api/XmlTypeLoader.java <==
/*
 * Copyright (c) MuleSoft, Inc.  All rights reserved.  http://www.mulesoft.com
 * The software in this package is published under the terms of the CPAL v1.0
 * license, a copy of which has been included with this distribution in the
 * LICENSE.txt file.
 */
package org.mule.metadata.xml.api;

import static java.util.stream.Collectors.joining;
import static org.mule.metadata.api.utils.MetadataTypeUtils.addTypeAlias;
import static org.mule.metadata.api.utils.MetadataTypeUtils.isNotNull;
import static org.mule.metadata.xml.internal.utils.SchemaHelper.getDefaultValue;

import org.mule.apache.xerces.impl.dv.xs.XSSimpleTypeDecl;
import org.mule.apache.xerces.impl.xs.XSComplexTypeDecl;
import org.mule.apache.xerces.impl.xs.XSElementDecl;
import org.mule.apache.xerces.xs.XSAnnotation;
import org.mule.apache.xerces.xs.XSAttributeDeclaration;
import org.mule.apache.xerces.xs.XSAttributeUse;
import org.mule.apache.xerces.xs.XSComplexTypeDefinition;
import org.mule.apache.xerces.xs.XSElementDeclaration;
import org.mule.apache.xerces.xs.XSModel;
import org.mule.apache.xerces.xs.XSModelGroup;
import org.mule.apache.xerces.xs.XSObjectList;
import org.mule.apache.xerces.xs.XSParticle;
import org.mule.apache.xerces.xs.XSTerm;
import org.mule.apache.xerces.xs.XSTypeDefinition;
import org.mule.apache.xerces.xs.XSWildcard;
import org.mule.metadata.api.TypeLoader;
import org.mule.metadata.api.builder.AttributeFieldTypeBuilder;
import org.mule.metadata.api.builder.BaseTypeBuilder;
import org.mule.metadata.api.builder.ObjectFieldTypeBuilder;
import org.mule.metadata.api.builder.ObjectKeyBuilder;
import org.mule.metadata.api.builder.ObjectTypeBuilder;
import org.mule.metadata.api.builder.TypeBuilder;
import org.mule.metadata.api.builder.UnionTypeBuilder;
import org.mule.metadata.api.model.MetadataFormat;
import org.mule.metadata.api.model.MetadataType;
import org.mule.metadata.xml.api.handler.HandlerManager;
import org.mule.metadata.xml.api.utils.SchemaHelper;

import javax.xml.namespace.QName;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Function;

import org.w3c.dom.TypeInfo;

public class XmlTypeLoader implements TypeLoader {

  public static final String TEXT_CONTENT_FIELD_NAME = "text()";
  public static MetadataFormat XML = MetadataFormat.XML;
  private static final String ROOT_ELEMENT_MARK = "#root:";

  private HandlerManager handlerManager;
  private ModelFactory modelFactory;
  private Map<QName, TypeBuilder> types;

  /**
   * Creates a new instance based on a group of schemas.
   *
   * @param schemaCollector an {@link SchemaCollector} with the schemas.
   */
  public XmlTypeLoader(SchemaCollector schemaCollector) {
    this(ModelFactory.fromSchemas(schemaCollector.collect()));
  }

  /**
   * Creates a new instance based on an already constructed {@link ModelFactory}.
   *
   * @param modelFactory a {@link ModelFactory} instance.
   */
  public XmlTypeLoader(ModelFactory modelFactory) {
    this.modelFactory = modelFactory;
    this.types = new HashMap<>();
    this.handlerManager = new HandlerManager();
  }

  @Override
  public Optional<MetadataType> load(String typeIdentifier) {
    return load(typeIdentifier, null);
  }

  @Override
  public Optional<MetadataType> load(String typeIdentifier, String typeAlias) {
    return load(typeIdentifier != null ? QName.valueOf(typeIdentifier) : null, typeAlias);
  }

  public Optional<MetadataType> load(QName qName, String typeAlias) {
    try {
      final XSModel model = modelFactory.getModel();
      QName rootElement = qName == null ? modelFactory.getRootElementName().orElse(null) : qName;
      if (model != null) {

        final XSElementDeclaration elementDeclaration =
            model.getElementDeclaration(rootElement.getLocalPart(), rootElement.getNamespaceURI());

        if (elementDeclaration == null) {
          return Optional.empty();
        }
        final BaseTypeBuilder rootBuilder = BaseTypeBuilder.create(XML);
        // Create root object with the root name
        final ObjectTypeBuilder rootObject = rootBuilder.objectType();
        rootObject.id(ROOT_ELEMENT_MARK + rootElement.toString());
        modelFactory.getExample().ifPresent(rootObject::with);
        final ObjectFieldTypeBuilder field = rootObject.addField().required();
        final ObjectKeyBuilder key = field.key(rootElement);

        processAttributes(elementDeclaration, key);
        final BaseTypeBuilder valueBuilder = field.value();
        processValue(elementDeclaration, valueBuilder);
        addTypeAlias(rootObject, typeAlias);
        return Optional.ofNullable(rootBuilder.build());
      } else {
        return Optional.empty();
      }
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  private void processValue(XSElementDeclaration elementDeclaration, BaseTypeBuilder valueBuilder) {
    final XSTypeDefinition typeDefinition = getXsTypeDefinition(elementDeclaration);
    if (typeDefinition instanceof XSComplexTypeDecl) {
      final XSComplexTypeDecl complexType = (XSComplexTypeDecl) typeDefinition;
      final Optional<Short> groupType = getGroupType(complexType);
      if (groupType.isPresent() && groupType.get().equals(XSModelGroup.COMPOSITOR_CHOICE)) {
        processUnionType(valueBuilder, complexType);
      } else {
        if (XSComplexTypeDefinition.CONTENTTYPE_SIMPLE == complexType.getContentType()) {
          valueBuilder.stringType();
        } else {
          final ObjectTypeBuilder objectType = valueBuilder.objectType();
          if (XSComplexTypeDefinition.CONTENTTYPE_MIXED == complexType.getContentType()) {
            objectType.addField().key(TEXT_CONTENT_FIELD_NAME).value().stringType();
          }
          final Optional<QName> typeName = getTypeName(typeDefinition);
          if (typeName.isPresent()) {
            objectType.id(typeName.get().toString());
            types.put(typeName.get(), objectType);
          }
          processAnnotations(complexType, objectType);
          processElements(complexType, objectType);
        }
      }
    } else if (typeDefinition instanceof XSSimpleTypeDecl) {
      handlerManager.handle((XSSimpleTypeDecl) typeDefinition, getDefaultValue(elementDeclaration), valueBuilder);
    }
  }

  private XSTypeDefinition getXsTypeDefinition(XSElementDeclaration elementDeclaration) {
    if (elementDeclaration instanceof XSElementDecl) {
      org.mule.apache.xerces.xni.QName unresolved = ((XSElementDecl) elementDeclaration).fUnresolvedTypeName;
      if (unresolved != null) {
        XSTypeDefinition definition = modelFactory.getModel().getTypeDefinition(unresolved.localpart, unresolved.uri);
        if (definition != null) {
          return definition;
        }
      }
    }
    return elementDeclaration.getTypeDefinition();
  }

  private void processUnionType(BaseTypeBuilder typeBuilder, XSComplexTypeDecl complexType) {
    final UnionTypeBuilder unionType = typeBuilder.unionType();
    final Optional<QName> unionTypeName = getTypeName(complexType);
    unionTypeName.ifPresent(qName -> types.put(qName, unionType));
    final XSParticle particle = complexType.getParticle();
    if (particle != null) {
      final XSTerm term = particle.getTerm();
      if (term instanceof XSModelGroup) {
        final XSModelGroup xsModelGroup = (XSModelGroup) term;
        final XSObjectList particles = xsModelGroup.getParticles();
        for (Object xsObject : particles) {
          if (xsObject instanceof XSParticle) {
            final XSParticle part = (XSParticle) xsObject;
            final XSTerm element = part.getTerm();
            final ObjectTypeBuilder objectType = unionType.of().objectType();
            if (element instanceof XSElementDeclaration) {
              final XSElementDeclaration elementDeclaration = (XSElementDeclaration) element;
              final String name = element.getName();
              final String namespace = element.getNamespace();
              final ObjectFieldTypeBuilder field = objectType.addField();
              final ObjectKeyBuilder key = field.key(new QName(namespace, name));
              processAttributes(elementDeclaration, key);
              field.required(part.getMinOccurs() > 0);
              final XSTypeDefinition typeDefinition = ((XSElementDeclaration) element).getTypeDefinition();
              final Optional<QName> typeName = getTypeName(typeDefinition);
              if (typeName.isPresent() && types.containsKey(typeName.get())) {
                field.value(types.get(typeName.get()));
              } else {
                final BaseTypeBuilder fieldValue = field.value();
                processValue(elementDeclaration, fieldValue);
              }
            } else {
              processElements(part, objectType);
            }
          }
        }
      }
    }
  }

  private Optional<Short> getGroupType(XSComplexTypeDecl complexType) {
    final XSParticle particle = complexType.getParticle();
    if (particle != null) {
      final XSTerm term = particle.getTerm();
      if (term instanceof XSModelGroup) {
        return Optional.of(((XSModelGroup) term).getCompositor());
      }
    }
    return Optional.empty();
  }

  private void processAnnotations(XSComplexTypeDecl complexType, ObjectTypeBuilder objectType) {
    final XSObjectList annotations = complexType.getAnnotations();
    for (Object annotation : annotations) {
      final XSAnnotation xsAnnotation = (XSAnnotation) annotation;
      final Collection<SchemaHelper.XmlDoc> xmlDocs = SchemaHelper.getDocumentation(xsAnnotation.getAnnotationString());
      String content = xmlDocs.stream().map(doc -> doc.getContent()).collect(joining("\n "));
      Optional<String> language = xmlDocs.stream().map(doc -> doc.getLang().orElse(null)).filter(Objects::nonNull).findFirst();
      objectType.description(language.orElse(null), content);
    }
  }

  private void processAttributes(XSElementDeclaration elementDeclaration, ObjectKeyBuilder keyBuilder) {
    final XSTypeDefinition typeDefinition = elementDeclaration.getTypeDefinition();
    if (typeDefinition instanceof XSComplexTypeDecl) {
      final XSObjectList attributeUses = ((XSComplexTypeDefinition) typeDefinition).getAttributeUses();
      for (Object attributeUse : attributeUses) {
        if (attributeUse instanceof XSAttributeUse) {
          final XSAttributeUse xsAttributeUse = (XSAttributeUse) attributeUse;
          final XSAttributeDeclaration attrDeclaration = xsAttributeUse.getAttrDeclaration();
          final String localName = attrDeclaration.getName();
          final String namespace = attrDeclaration.getNamespace();
          final AttributeFieldTypeBuilder attributeField = keyBuilder.addAttribute();
          attributeField.name(new QName(namespace, localName));
          attributeField.required(xsAttributeUse.getRequired());
          handlerManager.handle((XSSimpleTypeDecl) attrDeclaration.getTypeDefinition(),
                                getDefaultValue(xsAttributeUse),
                                attributeField.value());
        }
      }
    }
  }

  private void processElements(XSComplexTypeDecl complexType, ObjectTypeBuilder objectType) {
    processElements(complexType.getParticle(), objectType);
  }

  private void processElements(XSParticle particle, ObjectTypeBuilder objectType) {
    processElements(particle, objectType, (part) -> part.getMinOccurs() > 0);
  }

  private void processElements(XSParticle particle, ObjectTypeBuilder objectType,
                               Function<XSParticle, Boolean> requiredSupplier) {
    if (particle == null) {
      return;
    }

    final XSTerm term = particle.getTerm();
    if (term instanceof XSModelGroup) {
      final XSModelGroup xsModelGroup = (XSModelGroup) term;
      objectType.ordered(xsModelGroup.getCompositor() == XSModelGroup.COMPOSITOR_SEQUENCE);
      final XSObjectList particles = xsModelGroup.getParticles();
      for (Object xsObject : particles) {
        if (xsObject instanceof XSParticle) {
          final XSParticle part = (XSParticle) xsObject;
          final XSTerm element = part.getTerm();
          if (element instanceof XSElementDeclaration) {
            final XSElementDeclaration xsElementDeclaration = (XSElementDeclaration) element;
            final String name = element.getName();
            final String namespace = element.getNamespace();
            final ObjectFieldTypeBuilder field = objectType.addField();
            processAttributes(xsElementDeclaration, field.key(new QName(namespace, name)));
            field.required(requiredSupplier.apply(part));
            field.repeated(isRepeated(part));

            Number min = null;
            Number max = null;

            if (part.getMinOccurs() > 1) {
              min = part.getMinOccurs();
            }

            if (part.getMaxOccurs() > 1) {
              max = part.getMaxOccurs();
            }

            if (isNotNull(min) || isNotNull(max)) {
              field.occurrence(min, max);
            }

            final XSTypeDefinition typeDefinition = ((XSElementDeclaration) element).getTypeDefinition();
            final Optional<QName> typeName = getTypeName(typeDefinition);
            if (typeName.isPresent() && types.containsKey(typeName.get())) {
              field.value(types.get(typeName.get()));
            } else {
              final BaseTypeBuilder fieldValue = field.value();
              processValue(xsElementDeclaration, fieldValue);
            }
          } else if (element instanceof XSWildcard) {
            if (((XSWildcard) element).getConstraintType() == XSWildcard.NSCONSTRAINT_ANY) {
              objectType.open();
            }
          } else if (element instanceof XSModelGroup) {
            processElements(part, objectType, (xsParticle) -> false);
          }
        }
      }
    }
  }

  private Optional<QName> getTypeName(XSTypeDefinition typeDefinition) {
    Optional<QName> typeQname = Optional.empty();
    if (typeDefinition instanceof TypeInfo) {
      final TypeInfo definition = (TypeInfo) typeDefinition;
      if (definition.getTypeName() != null) {
        typeQname = Optional.of(new QName(definition.getTypeNamespace(), definition.getTypeName()));
      }
    }
    return typeQname;
  }

  private boolean isRepeated(XSParticle part) {
    return part.getMaxOccursUnbounded() || part.getMaxOccurs() > 1 || part.getMinOccurs() > 1;
  }

}
